import gettext
import locale
import string
import sys
from math import floor

RGB = "rgb"
HSB = "hsb"
CMYK = "cmyk"

BUTT = "butt"
ROUND = "round"
SQUARE = "square"
BEVEL = "bevel"
MITER = "miter"

APP = "shoebot"
DIR = sys.prefix + "/share/shoebot/locale"
locale.setlocale(locale.LC_ALL, "")
gettext.bindtextdomain(APP, DIR)
gettext.textdomain(APP)
_ = gettext.gettext


class Color(object):
    """Represents a single color.

    Attributes (RGB and HSL) are values between 0 and 1

    This stores color values as a list of 4 floats (RGBA) in a 0-1 range.

    The value can come in the following flavours:
    - v
    - (v)
    - (v,a)
    - (r,g,b)
    - (r,g,b,a)
    - #RRGGBB
    - RRGGBB
    - #RRGGBBAA
    - RRGGBBAA
    """

    def __init__(self, *args, **kwargs):
        if isinstance(args, Color):
            args = args.copy()

        color_range = float(kwargs.get("color_range", 1.0))
        mode = kwargs.get("mode", "rgb").lower()

        # Values are supplied as a tuple.
        if len(args) == 1 and isinstance(args[0], tuple):
            args = args[0]

        # No values or None, transparent black.
        if len(args) == 0 or (len(args) == 1 and args[0] is None):
            self.r, self.g, self.b, self.a = 0, 0, 0, 0

        # One value, another color object.
        elif len(args) == 1 and isinstance(args[0], Color):
            self.r, self.g, self.b, self.a = args[0].r, args[0].g, args[0].b, args[0].a

        # One value, a hexadecimal string.
        elif len(args) == 1 and isinstance(args[0], str):
            r, g, b, args = hex2rgb(args[0])
            self.r, self.g, self.b, self.a = r, g, b, args

        # One value, grayscale.
        elif len(args) == 1:
            gs = args[0]
            self.r, self.g, self.b, self.a = (
                gs / color_range,
                gs / color_range,
                gs / color_range,
                1,
            )

        # Two values, grayscale and alpha OR hex and alpha.
        elif len(args) == 2:
            if isinstance(args[0], str):
                alpha = args[1]
                r, g, b, args = hex2rgb(args[0])
                self.r, self.g, self.b, self.a = r, g, b, alpha
            else:
                gs, a = args[0], args[1]
                self.r, self.g, self.b, self.a = (
                    gs / color_range,
                    gs / color_range,
                    gs / color_range,
                    a / color_range,
                )

        # Three to five parameters, either RGB, RGBA, HSB, HSBA,
        # depending on the mode parameter.
        elif len(args) >= 3:
            alpha = 1
            if len(args) > 3:
                alpha = args[-1] / color_range

            if mode == "rgb":
                self.r, self.g, self.b, self.a = (
                    args[0] / color_range,
                    args[1] / color_range,
                    args[2] / color_range,
                    alpha,
                )
            elif mode == "hsb":
                self.h, self.s, self.brightness, self.a = (
                    args[0] / color_range,
                    args[1] / color_range,
                    args[2] / color_range,
                    alpha,
                )

    def __repr__(self):
        return "%s(%.3f, %.3f, %.3f, %.3f)" % (
            self.__class__.__name__,
            self.r,
            self.g,
            self.b,
            self.a,
        )

    @property
    def red(self):
        # Added
        return self.r

    @property
    def green(self):
        # Added
        return self.g

    @property
    def blue(self):
        # Added
        return self.a

    @property
    def alpha(self):
        # Added
        return self.a

    @property
    def data(self):
        # Added
        return [self.r, self.g, self.b, self.a]

    def copy(self):
        return tuple(self.data)

    def _update_rgb(self, r, g, b):
        self.__dict__["__r"] = r
        self.__dict__["__g"] = g
        self.__dict__["__b"] = b

    def _update_hsb(self, h, s, b):
        self.__dict__["__h"] = h
        self.__dict__["__s"] = s
        self.__dict__["__brightness"] = b

    def _hasattrs(self, list):
        for a in list:
            if a not in self.__dict__:
                return False
        return True

    # added
    def __getitem__(self, index):
        return (self.r, self.g, self.b, self.a)[index]

    def __iter__(self):
        for i in range(len(self.data)):
            yield self.data[i]

    def __div__(self, other):
        value = float(other)
        return (
            self.red / value,
            self.green / value,
            self.blue / value,
            self.alpha / value,
        )

    # end added

    def __setattr__(self, a, v):
        if a in ["a", "alpha"]:
            self.__dict__["__" + a[0]] = max(0, min(v, 1))

        # RGB changes, update HSB accordingly.
        elif a in ["r", "g", "b", "red", "green", "blue"]:
            self.__dict__["__" + a[0]] = max(0, min(v, 1))
            if self._hasattrs(("__r", "__g", "__b")):
                r, g, b = (
                    self.__dict__["__r"],
                    self.__dict__["__g"],
                    self.__dict__["__b"],
                )
                self._update_hsb(*rgb2hsb(r, g, b))

        # HSB changes, update RGB accordingly.
        elif a in ["h", "s", "hue", "saturation", "brightness"]:
            if a != "brightness":
                a = a[0]
            if a == "h":
                v = min(v, 0.99999999)
            self.__dict__["__" + a] = max(0, min(v, 1))
            if self._hasattrs(("__h", "__s", "__brightness")):
                r, g, b = hsb2rgb(
                    self.__dict__["__h"],
                    self.__dict__["__s"],
                    self.__dict__["__brightness"],
                )
                self._update_rgb(r, g, b)

        else:
            self.__dict__[a] = v

    def __getattr__(self, a):

        """Available properties:
        r, g, b, a or red, green, blue, alpha
        h, s or hue, saturation, brightness

        """
        if a in self.__dict__:
            return a
        elif a == "brightness":
            return self.__dict__["__brightness"]
        elif a in [
            "a",
            "alpha",
            "r",
            "g",
            "b",
            "red",
            "green",
            "blue",
            "h",
            "s",
            "hue",
            "saturation",
        ]:
            return self.__dict__["__" + a[0]]
        elif a in [
            "a",
            "alpha",
            "r",
            "g",
            "b",
            "red",
            "green",
            "blue",
            "h",
            "s",
            "hue",
            "saturation",
        ]:
            return self.__dict__["__" + a[0]]

        raise AttributeError(
            "'" + str(self.__class__) + "' object has no attribute '" + a + "'"
        )


class ColorMixin(object):
    """
    Mixin class for color support.
    Adds fill, stroke and blending mode attributes to the class.

    Setting color attributes, will convert them to Color instances, allowing
    them to be specfied in other ways, such as   fill="#123456"
    """

    def __init__(
        self,
        fill=None,
        fillrule=None,
        stroke=None,
        strokewidth=None,
        strokecap=None,
        strokejoin=None,
        strokedash=None,
        dashoffset=0,
        blendmode=None,
        **kwargs
    ):
        self._set_fill(fill)
        self._set_stroke(stroke)
        self._set_strokewidth(strokewidth)

        self.fillrule = self._fillrule = fillrule
        self.strokecap = self._strokecap = strokecap
        self.strokejoin = self._strokejoin = strokejoin
        self.strokedash = self._strokedash = strokedash
        self.dashoffset = self._dashoffset = dashoffset
        self.blendmode = self._blendmode = blendmode

    def _get_fill(self):
        return self._fillcolor

    def _set_fill(self, *args):
        if len(args) == 1:
            if args[0] is None:
                self._fillcolor = None
                return
            elif isinstance(args[0], Color):
                self._fillcolor = Color(args[0])
                return
        self._fillcolor = Color(mode="rgb", color_range=1, *args)

    fill = property(_get_fill, _set_fill)

    def _get_stroke(self):
        return self._strokecolor

    def _set_stroke(self, *args):
        if len(args) == 1:
            if args[0] is None:
                self._strokecolor = None
                return
            elif isinstance(args[0], Color):
                self._strokecolor = Color(args[0])
                return
        self._strokecolor = Color(mode="rgb", color_range=1, *args)

    stroke = property(_get_stroke, _set_stroke)

    def _get_strokewidth(self):
        return self._strokewidth

    def _set_strokewidth(self, strokewidth):
        self._strokewidth = max(strokewidth, 0.0001)

    strokewidth = property(_get_strokewidth, _set_strokewidth)


def hex2dec(hexdata):
    return int(string.atoi(hexdata, 16))


def dec2hex(d):
    return hex(d).split("x")[-1]


def parse_color(v, color_range=1):
    """Receives a colour definition and returns a (r,g,b,a) tuple.

    Accepts:
    - v
    - (v)
    - (v,a)
    - (r,g,b)
    - (r,g,b,a)
    - #RRGGBB
    - RRGGBB
    - #RRGGBBAA
    - RRGGBBAA

    Returns a (red, green, blue, alpha) tuple, with values ranging from
    0 to 1.

    The 'color_range' parameter sets the colour range in which the
    colour data values are specified (except in hexstrings).
    """

    # unpack one-element tuples, they show up sometimes
    while isinstance(v, (tuple, list)) and len(v) == 1:
        v = v[0]

    if isinstance(v, (int, float)):
        red = green = blue = v / color_range
        alpha = 1.0

    elif isinstance(v, Color):
        red, green, blue, alpha = v

    elif isinstance(v, (tuple, list)):
        # normalise values according to the supplied colour range
        # for this we make a list with the normalised data
        color = []
        for index in range(0, len(v)):
            color.append(v[index] / color_range)

        if len(color) == 1:
            red = green = blue = alpha = color[0]
        elif len(color) == 2:
            red = green = blue = color[0]
            alpha = color[1]
        elif len(color) == 3:
            red = color[0]
            green = color[1]
            blue = color[2]
            alpha = 1.0
        elif len(color) == 4:
            red = color[0]
            green = color[1]
            blue = color[2]
            alpha = color[3]

    elif isinstance(v, str):
        # got a hexstring: first remove hash character, if any
        v = v.strip("#")
        if len(v) == 6:
            # RRGGBB
            red = hex2dec(v[0:2]) / 255.0
            green = hex2dec(v[2:4]) / 255.0
            blue = hex2dec(v[4:6]) / 255.0
            alpha = 1.0
        elif len(v) == 8:
            red = hex2dec(v[0:2]) / 255.0
            green = hex2dec(v[2:4]) / 255.0
            blue = hex2dec(v[4:6]) / 255.0
            alpha = hex2dec(v[6:8]) / 255.0

    return red, green, blue, alpha


# Some generic color conversion algorithms used mainly by BaseColor outside of NodeBox.


def hex_to_rgb(hex):
    """Returns RGB values for a hex color string."""
    hex = hex.lstrip("#")
    if len(hex) < 6:
        hex += hex[-1] * (6 - len(hex))
    if len(hex) == 6:
        r, g, b = hex[0:2], hex[2:4], hex[4:]
        r, g, b = [int(n, 16) / 255.0 for n in (r, g, b)]
        a = 1.0
    elif len(hex) == 8:
        r, g, b, a = hex[0:2], hex[2:4], hex[4:6], hex[6:]
        r, g, b, a = [int(n, 16) / 255.0 for n in (r, g, b, a)]
    return r, g, b, a


hex2rgb = hex_to_rgb


def lab_to_rgb(l, a, b):
    """Converts CIE Lab to RGB components.

    First we have to convert to XYZ color space.
    Conversion involves using a white point,
    in this case D65 which represents daylight illumination.

    Algorithm adopted from:
    http://www.easyrgb.com/math.php

    """

    y = (l + 16) / 116.0
    x = a / 500.0 + y
    z = y - b / 200.0
    v = [x, y, z]
    for i in _range(3):
        if pow(v[i], 3) > 0.008856:
            v[i] = pow(v[i], 3)
        else:
            v[i] = (v[i] - 16 / 116.0) / 7.787

    # Observer = 2, Illuminant = D65
    x = v[0] * 95.047 / 100
    y = v[1] * 100.0 / 100
    z = v[2] * 108.883 / 100

    r = x * 3.2406 + y * -1.5372 + z * -0.4986
    g = x * -0.9689 + y * 1.8758 + z * 0.0415
    b = x * 0.0557 + y * -0.2040 + z * 1.0570
    v = [r, g, b]
    for i in _range(3):
        if v[i] > 0.0031308:
            v[i] = 1.055 * pow(v[i], 1 / 2.4) - 0.055
        else:
            v[i] = 12.92 * v[i]

    r, g, b = v[0], v[1], v[2]
    return r, g, b


lab2rgb = lab_to_rgb


def hsv_to_rgb(h, s, v):
    """Hue, saturation, brightness to red, green, blue.
    http://www.koders.com/python/fidB2FE963F658FE74D9BF74EB93EFD44DCAE45E10E.aspx
    Results will differ from the way NSColor converts color spaces.
    """

    if s == 0:
        return v, v, v

    h = h / (60.0 / 360)
    i = floor(h)
    f = h - i
    p = v * (1 - s)
    q = v * (1 - s * f)
    t = v * (1 - s * (1 - f))

    if i == 0:
        r, g, b = v, t, p
    elif i == 1:
        r, g, b = q, v, p
    elif i == 2:
        r, g, b = p, v, t
    elif i == 3:
        r, g, b = p, q, v
    elif i == 4:
        r, g, b = t, p, v
    else:
        r, g, b = v, p, q

    return r, g, b


hsv2rgb = hsb2rgb = hsb_to_rgb = hsv_to_rgb


def rgb_to_hsv(r, g, b):
    h = s = 0
    v = max(r, g, b)
    d = v - min(r, g, b)

    if v != 0:
        s = d / float(v)

    if s != 0:
        if r == v:
            h = 0 + (g - b) / d
        elif g == v:
            h = 2 + (b - r) / d
        else:
            h = 4 + (r - g) / d

    h = h * (60.0 / 360)
    if h < 0:
        h = h + 1.0

    return h, s, v


rgb2hsv = rgb2hsb = rgb_to_hsb = rgb_to_hsv


def rgba_to_argb(stringImage):
    tempBuffer = [None] * len(
        stringImage
    )  # Create an empty array of the same size as stringImage
    tempBuffer[0::4] = stringImage[2::4]
    tempBuffer[1::4] = stringImage[1::4]
    tempBuffer[2::4] = stringImage[0::4]
    tempBuffer[3::4] = stringImage[3::4]
    stringImage = "".join(tempBuffer)
    return stringImage


def parse_hsb_color(v, color_range=1):
    if isinstance(v, str):
        # hexstrings aren't hsb
        return parse_color(v)
    hue, saturation, brightness, alpha = parse_color(v, color_range)
    red, green, blue = hsv_to_rgb(hue, saturation, brightness)
    return red, green, blue, alpha



# TODO, Move 'bot' out of here, push responsibility back to 'Nodebox' and other grammars,
#       Enabling seperation of BezierPath etc

from shoebot.core.backend import cairo

# Alignments
CENTER = "center"
CORNER = "corner"
CORNERS = "corners"

STATES = {
    "transform": "_transform",
    "fillcolor": "_fillcolor",
    "fillrule": "_fillrule",
    "strokecolor": "_strokecolor",
    "strokewidth": "_strokewidth",
    "strokecap": "_strokecap",
    "strokejoin": "_strokejoin",
    "strokedash": "_strokedash",
    "dashoffset": "_dashoffset",
    "blendmode": "_blendmode",
    "align": "_align",
    "fontsize": "_fontsize",
    "lineheight": "_lineheight",
}


class Grob:
    """A GRaphic OBject is the base class for all DrawingPrimitives."""

    def __init__(self, bot):
        # Takes bot rather than canvas for compatibility with libraries - e.g. the colors library
        self._canvas = canvas = bot._canvas
        self._bot = bot
        self._set_mode(canvas.mode)
        self._transform = cairo.Matrix(*canvas.transform)

    def _set_mode(self, mode):
        """
        Sets call_transform_mode to point to the
        center_transform or corner_transform
        """
        if mode == CENTER:
            self._call_transform_mode = self._center_transform
        elif mode == CORNER:
            self._call_transform_mode = self._corner_transform
        else:
            raise ValueError("mode must be CENTER or CORNER")

    def _get_center(self):
        """Implementations must return the x, y of their center"""
        raise NotImplementedError()

    def _call_transform_mode(self):
        """
        This should never get called:
        set mode, changes the value of this to point

        corner_transform or center_transform
        """
        raise NotImplementedError("_call_transform_mode called without mode set!")

    def _center_transform(self, transform):
        """'
        Works like setupTransform of a version of java nodebox
        http://dev.nodebox.net/browser/nodebox-java/branches/rewrite/src/java/net/nodebox/graphics/Grob.java
        """
        dx, dy = self._get_center()
        t = cairo.Matrix()
        t.translate(dx, dy)
        t = transform * t
        t.translate(-dx, -dy)
        return t

    def _corner_transform(self, transform):
        """
        CORNER is the default, so we just return the transform
        """
        return transform

    def _deferred_render(self, render_func=None):
        """
        Pass a function to the canvas for deferred rendering,
        defaults to self._render
        """
        self._canvas.deferred_render(render_func or self._render)

    def _render(self, ctx):
        """For overriding by GRaphicOBjects"""
        raise NotImplementedError()

    def inheritFromContext(self, ignore=()):
        """
        Doesn't store exactly the same items as Nodebox for ease of implementation,
        it has enough to get the Nodebox Dentrite example working.
        """
        for canvas_attr, grob_attr in list(STATES.items()):
            if canvas_attr in ignore:
                continue
            setattr(self, grob_attr, getattr(self._bot._canvas, canvas_attr))



import linecache
import sys
import traceback


def simple_traceback(ex, source):
    """
    Format traceback, showing line number and surrounding source.
    """
    exc_type, exc_value, exc_tb = sys.exc_info()
    exc = traceback.format_exception(exc_type, exc_value, exc_tb)

    source_arr = source.splitlines()

    exc_location = exc[-2]
    exc_script_index = 0
    for i, err in enumerate(exc):
        if 'File "<string>",' in err:
            exc_script_index = i
            exc_location = exc[exc_script_index]
            break

    # Extract line number from traceback
    fn = exc_location.split(",")[0][8:-1]
    line_number = int(exc_location.split(", ")[1].lstrip("line "))

    err_msgs = []

    # Output code around the error
    err_where = exc[i].split(",")[1].strip()  # 'line 37 in blah"
    err_msgs.append("Error in the Shoebot script at %s:" % err_where)
    for i in range(max(0, line_number - exc_script_index), line_number):
        if fn == "<string>":
            line = source_arr[i]
        else:
            line = linecache.getline(fn, i + 1)
        err_msgs.append("%s: %s" % (i + 1, line.rstrip()))
    err_msgs.append("  %s^ %s" % (len(str(i)) * " ", exc[-1].rstrip()))

    err_msgs.append("")

    # traceback
    err_msgs.append(exc[0].rstrip())
    for err in exc[exc_script_index:]:
        err_msgs.append(err.rstrip())

    return "\n".join(err_msgs)



#!/usr/bin/env python3

# This file is part of Shoebot.
# Copyright (C) 2007-2009 the Shoebot authors
# See the COPYING file for the full license text.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#   Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
#   The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
#   WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
#   EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
#   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
#   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""
Drawbot and Nodebot are similar grammars, so they both inherit from Bot
"""

import sys
import os

from shoebot.core.backend import cairo
from shoebot.data import (
    BezierPath,
    EndClip,
    Color,
    Text,
    Variable,
    Image,
    ClippingPath,
    NUMBER,
    TEXT,
    BOOLEAN,
    BUTTON,
    ShoebotError,
    RGB,
    HSB,
    CENTER,
    CORNER,
    CORNERS,
)

from .grammar import Grammar

from pkg_resources import resource_filename, Requirement
from glob import glob
import random as r

import locale
import gettext

SBOT_ROOT = resource_filename(Requirement.parse("shoebot"), "")
APP = "shoebot"
DIR = sys.prefix + "/share/shoebot/locale"
locale.setlocale(locale.LC_ALL, "")
gettext.bindtextdomain(APP, DIR)
# gettext.bindtextdomain(APP)
gettext.textdomain(APP)
_ = gettext.gettext

LIB_DIRS = [
    os.path.join(SBOT_ROOT, "local", "share", "shoebot", "lib"),
    os.path.join(SBOT_ROOT, "lib"),
    os.path.join(SBOT_ROOT, "share", "shoebot", "lib"),
    os.path.join(sys.prefix, "share", "shoebot", "lib"),
]
for LIB_DIR in LIB_DIRS:
    if os.path.isdir(LIB_DIR):
        sys.path.append(LIB_DIR)

TOP_LEFT = 1
BOTTOM_LEFT = 2


class Bot(Grammar):
    """
    The Parts of the Grammar common to DrawBot, NodeBot and ShoeBot.
    """

    RGB = RGB
    HSB = HSB

    LEFT = "left"
    RIGHT = "right"

    CENTER = CENTER
    CORNER = CORNER
    CORNERS = CORNERS

    LEFT = "left"
    RIGHT = "right"
    JUSTIFY = "justify"

    NUMBER = NUMBER
    TEXT = TEXT
    BOOLEAN = BOOLEAN
    BUTTON = BUTTON

    inch = 72
    cm = 28.3465
    mm = 2.8346

    # Default mouse values
    MOUSEX = -1
    """The x-value of the mouse cursor coordinates."""
    MOUSEY = -1
    """The y-value of the mouse cursor coordinates."""
    mousedown = False
    """Is True if the mouse button is pressed."""

    # Default key values
    key = "-"
    keycode = 0
    keydown = False

    def __init__(self, canvas, namespace=None, vars=None):
        """
        :param canvas: Canvas implementation for output.
        :param namespace: Optionally specify a dict to inject as namespace
        :param vars: Optional dict containing initial values for variables
        """
        Grammar.__init__(self, canvas, namespace=namespace, vars=vars)
        canvas.set_bot(self)

        self._autoclosepath = True
        self._path = None

        if self._input_device:
            # Get constants like KEY_DOWN, KEY_LEFT
            for key_name, value in list(self._input_device.get_key_map().items()):
                self._namespace[key_name] = value
                setattr(self, key_name, value)

        self._canvas.size = None
        if isinstance(namespace, dict) and "FRAME" in namespace:
            try:
                self._frame = int(namespace["FRAME"])
            except ValueError:
                raise ValueError("Frame must be an integer.")
        else:
            self._frame = 1
        self._set_initial_defaults()  ### TODO Look at these

    def _set_initial_defaults(self):
        """Set the default values. Called at __init__ and at the end of run(),
        so that new draw loop iterations don't take up values left over by the
        previous one."""
        DEFAULT_WIDTH, DEFAULT_HEIGHT = self._canvas.DEFAULT_SIZE
        self.WIDTH = self._namespace.get("WIDTH", DEFAULT_WIDTH)
        self.HEIGHT = self._namespace.get("HEIGHT", DEFAULT_WIDTH)
        if "WIDTH" in self._namespace or "HEIGHT" in self._namespace:
            self.size(w=self._namespace.get("WIDTH"), h=self._namespace.get("HEIGHT"))

        self._transformmode = Bot.CENTER

        self._canvas.settings(
            fillcolor=self.color(0.2),
            fillrule=None,
            strokecolor=None,
            strokewidth=1.0,
            strokecap=None,
            strokejoin=None,
            strokedash=None,
            dashoffset=0,
            blendmode=None,
            background=self.color(1, 1, 1),
            fontfile="Sans",
            fontsize=16,
            align=Bot.LEFT,
            lineheight=1,
            tracking=0,
            underline=None,
            overline=None,
            underlinecolor=None,
            overlinecolor=None,
            hintstyle=None,
            hintmetrics=None,
            antialias=None,
            subpixelorder=None,
        )

    # Input GUI callbacks

    def _mouse_button_down(self, button):
        """GUI callback for mouse button down"""
        self._namespace["mousedown"] = True

    def _mouse_button_up(self, button):
        """GUI callback for mouse button up"""
        self._namespace["mousedown"] = self._input_device.mouse_down

    def _mouse_pointer_moved(self, x, y):
        """GUI callback for mouse moved"""
        self._namespace["MOUSEX"] = x
        self._namespace["MOUSEY"] = y

    def _key_pressed(self, key, keycode):
        """GUI callback for key pressed"""
        self._namespace["key"] = key
        self._namespace["keycode"] = keycode
        self._namespace["keydown"] = True

    def _key_released(self, key, keycode):
        """GUI callback for key released"""
        self._namespace["keydown"] = self._input_device.key_down

    # Functions for override #####

    # Classes #####

    def _makeInstance(self, clazz, args, kwargs):
        """Creates an instance of a class defined in this document.
        This method sets the context of the object to the current context."""
        inst = clazz(self, *args, **kwargs)
        return inst

    def _makeColorableInstance(self, clazz, args, kwargs):
        """
        Create an object, if fill, stroke or strokewidth
        is not specified, get them from the _canvas

        :param clazz:
        :param args:
        :param kwargs:
        :return:
        """
        kwargs = dict(kwargs)

        fill = kwargs.get("fill", self._canvas.fillcolor)
        if not isinstance(fill, Color):
            fill = Color(fill, mode="rgb", color_range=1)
        kwargs["fill"] = fill

        stroke = kwargs.get("stroke", self._canvas.strokecolor)
        if not isinstance(stroke, Color):
            stroke = Color(stroke, mode="rgb", color_range=1)
        kwargs["stroke"] = stroke

        kwargs["fillrule"] = kwargs.get("fillrule", self._canvas.fillrule)
        kwargs["strokewidth"] = kwargs.get("strokewidth", self._canvas.strokewidth)
        kwargs["strokecap"] = kwargs.get("strokecap", self._canvas.strokecap)
        kwargs["strokejoin"] = kwargs.get("strokejoin", self._canvas.strokejoin)
        kwargs["strokedash"] = kwargs.get("strokedash", self._canvas.strokedash)
        kwargs["dashoffset"] = kwargs.get("dashoffset", self._canvas.dashoffset)
        kwargs["blendmode"] = kwargs.get("blendmode", self._canvas.blendmode)
        inst = clazz(self, *args, **kwargs)
        return inst

    def EndClip(self, *args, **kwargs):
        return self._makeColorableInstance(EndClip, args, kwargs)

    def BezierPath(self, *args, **kwargs):
        return self._makeColorableInstance(BezierPath, args, kwargs)

    def ClippingPath(self, *args, **kwargs):
        return self._makeColorableInstance(ClippingPath, args, kwargs)

    def Rect(self, *args, **kwargs):
        return self._makeColorableInstance(Rect, args, kwargs)

    def Oval(self, *args, **kwargs):
        return self._makeColorableInstance(Oval, args, kwargs)

    def Ellipse(self, *args, **kwargs):
        return self._makeColorableInstance(Ellipse, args, kwargs)

    def Color(self, *args, **kwargs):
        return Color(*args, **kwargs)

    def Image(self, *args, **kwargs):
        return self._makeColorableInstance(Image, args, kwargs)

    def Text(self, *args, **kwargs):
        return self._makeColorableInstance(Text, args, kwargs)

    # Variables #####

    def var(self, name, type, default=None, min=0, max=255, value=None, step=None):
        v = Variable(
            name, type, default=default, min=min, max=max, value=value, step=step
        )
        return self._addvar(v)

    # Utility #####

    def color(self, *args):
        """
        :param args: color in a supported format.

        :return: Color object containing the color.
        """
        return self.Color(mode=self.color_mode, color_range=self.color_range, *args)

    choice = r.choice

    def random(self, v1=None, v2=None):
        # ipsis verbis from Nodebox
        if v1 is not None and v2 is None:
            if isinstance(v1, float):
                return r.random() * v1
            else:
                return int(r.random() * v1)
        elif v1 is not None and v2 is not None:
            if isinstance(v1, float) or isinstance(v2, float):
                start = min(v1, v2)
                end = max(v1, v2)
                return start + r.random() * (end - start)
            else:
                start = min(v1, v2)
                end = max(v1, v2) + 1
                return int(start + r.random() * (end - start))
        else:
            # No values means 0.0 -> 1.0
            return r.random()

    def grid(self, cols, rows, colSize=1, rowSize=1, shuffled=False):
        """Returns an iterator that contains coordinate tuples.

        This command can be used to quickly create grid-like structures.
        A common usage pattern is:

        .. code-block:: python

            for x, y in grid(10,10,12,12):
                rect(x,y, 10,10)
        """
        from random import shuffle

        rowRange = list(range(int(rows)))
        colRange = list(range(int(cols)))
        if shuffled:
            shuffle(rowRange)
            shuffle(colRange)
        for y in rowRange:
            for x in colRange:
                yield (x * colSize, y * rowSize)

    def files(self, path="*"):
        """Returns a list of files.

        Use wildcards to specify which files to pick, e.g. ``f =
        files('*.gif')``.

        :param path: wildcard to use in file list
        :return: list of file names
        """
        return glob(path)

    def snapshot(self, target=None, defer=None, autonumber=False):
        """Save the contents of current surface into a file or cairo surface/context.

        :param filename: Can be a filename or a Cairo surface.
        :param defer: When to snapshot, if set to True waits until the frame has finished rendering.
        :param autonumber: If true then a number will be appended to the filename.
        """
        if autonumber:
            file_number = self._frame
        else:
            file_number = None

        if isinstance(target, cairo.Surface):
            # snapshot to Cairo surface
            if defer is None:
                self._canvas.snapshot(target, defer)
                defer = False
            ctx = cairo.Context(target)
            # this used to be self._canvas.snapshot, but I couldn't make it work.
            # self._canvas.snapshot(target, defer)
            # TODO: check if this breaks when taking more than 1 snapshot
            self._canvas._drawqueue.render(ctx)
            return
        elif target is None:
            # If nothing specified, use a default filename from the script name
            script_file = self._namespace.get("__file__")
            if script_file:
                target = os.path.splitext(script_file)[0] + ".svg"
                file_number = True

        if target:
            # snapshot to file, target is a filename
            if defer is None:
                defer = True
            self._canvas.snapshot(target, defer=defer, file_number=file_number)
        else:
            raise ShoebotError(
                "Image not saved: no target, filename or default to save to."
            )

    def show(self, format="png", as_data=False):
        """Returns an Image object of the current surface. Used for displaying
        output in Jupyter notebooks. Adapted from the cairo-jupyter project."""

        from io import BytesIO

        b = BytesIO()

        if format == "png":
            from IPython.display import Image

            surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, self.WIDTH, self.HEIGHT)
            self.snapshot(surface)
            surface.write_to_png(b)
            b.seek(0)
            data = b.read()
            if as_data:
                return data
            else:
                return Image(data)
        elif format == "svg":
            from IPython.display import SVG

            surface = cairo.SVGSurface(b, self.WIDTH, self.HEIGHT)
            surface.restrict_to_version(cairo.SVGVersion.VERSION_1_2)
            surface.finish()
            b.seek(0)
            data = b.read()
            if as_data:
                return data
            else:
                return SVG(data)

    def ximport(self, libName):
        """
        Import Nodebox libraries.

        The libraries get _ctx, which provides
        them with the nodebox API.

        :param libName: Library name to import
        """
        # from Nodebox
        lib = __import__(libName)
        self._namespace[libName] = lib
        lib._ctx = self
        return lib

    # Core functions ####

    def size(self, w=None, h=None):
        """Set the canvas size

        Only the first call will actually be effective.

        :param w: Width
        :param h: height
        """

        if not w:
            w = self._canvas.width
        if not h:
            h = self._canvas.height
        if not w and not h:
            return self._canvas.width, self._canvas.height

        # FIXME: Updating in all these places seems a bit hacky
        w, h = self._canvas.set_size((w, h))
        self._namespace["WIDTH"] = w
        self._namespace["HEIGHT"] = h
        self.WIDTH = w  # Added to make evolution example work
        self.HEIGHT = h  # Added to make evolution example work

    def speed(self, framerate=None):
        """Set animation framerate.

        :param framerate: Frames per second to run bot.
        :return: Current framerate of animation.
        """
        if framerate is not None:
            self._speed = framerate
            self._dynamic = True
        else:
            return self._speed

    @property
    def FRAME(self):
        return self._frame



from math import hypot, degrees, atan2, sqrt, pow, radians, sin, cos


def distance(x1, y1, x2, y2):
    return hypot(x2 - x1, y2 - y1)


def angle(x0, y0, x1, y1):
    # from nodebox1
    a = degrees(atan2(y1 - y0, x1 - x0))
    return a


def coordinates(x0, y0, distance, angle):
    # from nodebox1
    x1 = x0 + cos(radians(angle)) * distance
    y1 = y0 + sin(radians(angle)) * distance
    return x1, y1


def reflect(x0, y0, x1, y1, d=1.0, a=180):
    # from nodebox1
    d *= distance(x0, y0, x1, y1)
    a += angle(x0, y0, x1, y1)
    x, y = coordinates(x0, y0, d, a)
    return x, y



import atexit
import threading
try:
    import numpy as np
except ImportError:
    print("The Audio library requires the numpy module (pip install numpy).")
    raise
try:
    from pysoundcard import InputStream
    from pysoundcard import device_info
except ImportError:
    print("The Audio library requires the pysoundcard module (pip install pysoundcard).")
    raise
try:
    from fuzzywuzzy import fuzz
except ImportError:
    print("The Audio library requires the fuzzywuzzy module (pip install fuzzywuzzy).")
    raise


BUFF_LENGTH = 1024
CHANNELS = 2
NUM_SAMPLES = 512


def fft_bandpassfilter(data, fs, lowcut, highcut):
    """
    http://www.swharden.com/blog/2009-01-21-signal-filtering-with-python/#comment-16801
    """
    fft = np.fft.fft(data)
    # n = len(data)
    # timestep = 1.0 / fs
    # freq = np.fft.fftfreq(n, d=timestep)
    bp = fft.copy()

    # Zero out fft coefficients
    # bp[10:-10] = 0

    # Normalise
    # bp *= real(fft.dot(fft))/real(bp.dot(bp))

    bp *= fft.dot(fft) / bp.dot(bp)

    # must multipy by 2 to get the correct amplitude
    ibp = 12 * np.fft.ifft(bp)
    return ibp

    # for i in range(len(bp)):
    #     if freq[i] >= highcut or freq[i] < lowcut:
    #         bp[i] = 0

    # ibp = np.fft.ifft(bp)
    # ibp = 2.0 * np.fft.ifft(bp)
    # print ibp
    # return ibp


def flatten_fft(scale=1.0):
    """
    Produces a nicer graph, I'm not sure if this is correct
    """
    _len = len(audio.spectrogram)
    for i, v in enumerate(audio.spectrogram):
        yield scale * (i * v) / _len


def scaled_fft(fft, scale=1.0):
    """
    Produces a nicer graph, I'm not sure if this is correct
    """
    data = np.zeros(len(fft))
    for i, v in enumerate(fft):
        data[i] = scale * (i * v) / NUM_SAMPLES

    return data


def triple(spectrogram):
    # c = spectrogram.copy()
    # c.resize(3, 255 / 3)
    # return c
    bass = spectrogram[0:85]
    mid = spectrogram[85:-85]
    treble = spectrogram[-85:-1]

    return bass, mid, treble


def fuzzydevices(match='', min_ratio=30):
    device_ratios = []
    for device in device_info():
        ratio = fuzz.partial_ratio(match, device['name'])
        if ratio > min_ratio:
            device_ratios.append((ratio, device))

    for ratio, device in sorted(device_ratios, key=lambda ratio_device: (ratio_device[0])):
        yield device


def firstfuzzydevice(match=''):
    devices = list(fuzzydevices(match, 0))
    return devices[0]


class AudioException(Exception):
    def __init__(self, msg):
        Exception.__init__(self)
        self.msg = msg

    def __str__(self):
        return self.msg


class AudioThread(threading.Thread):
    def __init__(self):
        # TODO - This could be a bit less dumb, and only run
        #        bandpass (or other) filters if they are actually used
        super(self.__class__, self).__init__()

        self.daemon = True

        self.streams = {}
        self.running = False

        self._spectrogram = np.zeros(NUM_SAMPLES)
        self._bandpassed = np.zeros(NUM_SAMPLES)

    def settings(self, **kwargs):
        if self.running:
            raise AudioException('Audio is already running')

    def run(self):
        with InputStream(samplerate=44100, blocksize=16) as s:
            while self.running:
                vec = s.read(NUM_SAMPLES)
                # Downsample to mono
                mono_vec = vec.sum(-1) / float(s.channels)
                self._spectrogram = np.fft.fft(mono_vec)
                self._bandpassed = fft_bandpassfilter(self.spectrogram, 44010, 100, 20)

    def autostart(self):
        if not self.running:
            self.running = True
            self.start()
            atexit.register(self.quit)

    def quit(self):
        """
        Shutdown the audio thread
        """
        if self.running:
            self.running = False
            self.join()

    @property
    def spectrogram(self):
        self.autostart()
        return self._spectrogram

    @property
    def bandpassed(self):
        self.autostart()
        return self._bandpassed


audio = AudioThread()



from random import random
from math import pi, sin, cos
from math import sqrt


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y


##### GRAPH LAYOUT ###################################################################################


class layout(object):
    """
    Graph visualizer that calculates relative node positions.
    """

    def __init__(self, graph, iterations=1000):
        self.type = None
        self.graph = graph
        self.i = 0
        self.n = iterations

        self.__bounds = None

    def copy(self, graph):
        """
        Returns a copy of the layout for the given graph.
        """
        l = self.__class__(graph, self.n)
        l.i = 0
        return l

    def prepare(self):
        for n in self.graph.nodes:
            n.vx = 0
            n.vy = 0
            n.force = Point(0, 0)

    def _bounds(self):
        min = Point(float("inf"), float("inf"))
        max = Point(float("-inf"), float("-inf"))
        for n in self.graph.nodes:
            if n.vx < min.x:
                min.x = n.vx
            if n.vy < min.y:
                min.y = n.vy
            if n.vx > max.x:
                max.x = n.vx
            if n.vy > max.y:
                max.y = n.vy

        return (min, max)

    bounds = property(_bounds)

    def _get_done(self):
        return self.i >= self.n

    done = property(_get_done)

    def iterate(self):
        self.i += 1
        return self.done

    def solve(self):
        while not self.done:
            self.iterate()

    def reset(self):
        self.i = 0

    def refresh(self):
        self.i = self.n / 2


##### GRAPH CIRCLE LAYOUT ############################################################################


class circle_layout(layout):
    """
    Simple layout with nodes arranged on one or more circles.
    """

    def __init__(self, graph, iterations=100):
        layout.__init__(self, graph, iterations)
        self.type = "circle"

        self.r = 8  # outer circle radius
        self.c = 2  # number of circles
        self.a = pi / 2  # starting angle

    def _get_orbits(self):
        return self.c

    def _set_orbits(self, v):
        self.c = v

    orbits = property(_get_orbits, _set_orbits)

    def copy(self, graph):
        l = layout.copy(self, graph)
        l.r = self.r
        return l

    def iterate(self):
        if len(self.graph.nodes) == 1:
            return

        # Nodes are sorted by betweenness centrality.
        # Node with a high centrality are on the inner circles.
        # There are logarithmically more nodes on the outer shells.
        circles = []
        nodes = self.graph.nodes_by_traffic(treshold=-1)
        for i in range(self.c):
            t = 1.0 / (self.c - i) ** 2
            slice = int(t * len(self.graph.nodes))
            slice = max(1, slice)
            circles.append(nodes[:slice])
            nodes = nodes[slice:]

        node_radius = self.graph.nodes[0].__class__(None).r

        i = 0
        for circle in circles:
            i += 1

            # Circle radii expand each iteration.
            # Inner circles have a smaller radius.
            r = self.r * sin(pi / 2 * float(self.i) / self.n)
            r *= float(i) / len(circles)

            # Calculate circle circumference.
            # Node diameter / circumference determine how many nodes fit on the shell.
            C = self.r * self.graph.d * 2 * pi * float(i) / len(circles)
            s = node_radius * 2 / C * 2

            a = self.a
            t = min(2 * pi * s, 2 * pi / len(circle))
            for n in circle:
                n.vx = r * cos(a)
                n.vy = r * sin(a)
                a += t

        # r = self.r * sin(float(self.i) / self.n * pi/2)
        # a = 0
        # i = pi*2 / len(self.graph.nodes)
        # for n in self.graph.nodes:
        #    n.vx = r * cos(a)
        #    n.vy = r * sin(a)
        #    a += i

        layout.iterate(self)

    def solve(self):
        self.i = self.n
        self.iterate()


##### GRAPH SPRING LAYOUT ############################################################################


class spring_layout(layout):
    """
    A force-based layout in which edges are regarded as springs.
    http://snipplr.com/view/1950/graph-javascript-framework-version-001/
    """

    def __init__(self, graph, iterations=1000):
        layout.__init__(self, graph, iterations)
        self.type = "spring"

        self.k = 2  # force strength
        self.m = 0.01  # force multiplier
        self.w = 15  # edge weight multiplier
        self.d = 0.5  # maximum vertex movement
        self.r = 15  # maximum repulsive force radius

    def tweak(self, k=2, m=0.01, w=15, d=0.5, r=15):
        self.k = k
        self.m = (m,)
        self.w = w
        self.d = d
        self.r = r

    def _get_force(self):
        return self.m

    def _set_force(self, v):
        self.m = v

    force = property(_get_force, _set_force)

    def _get_repulsion(self):
        return self.r

    def _set_repulsion(self, v):
        self.r = v

    repulsion = property(_get_repulsion, _set_repulsion)

    def copy(self, graph):

        l = layout.copy(self, graph)
        l.k, l.m, l.d, l.r = self.k, self.m, self.d, self.r
        return l

    def iterate(self):
        # Forces on all nodes due to node-node repulsions.
        for i in range(len(self.graph.nodes)):
            n1 = self.graph.nodes[i]
            for j in range(i + 1, len(self.graph.nodes)):
                n2 = self.graph.nodes[j]
                self._repulse(n1, n2)

        # Forces on nodes due to edge attractions.
        for e in self.graph.edges:
            self._attract(e.node1, e.node2, self.w * e.weight, 1.0 / e.length)

        # Move by given force.
        for n in self.graph.nodes:
            vx = max(-self.d, min(self.m * n.force.x, self.d))
            vy = max(-self.d, min(self.m * n.force.y, self.d))
            n.vx += vx
            n.vy += vy
            n.force.x = 0
            n.force.y = 0

        return layout.iterate(self)

    def _distance(self, n1, n2):
        dx = n2.vx - n1.vx
        dy = n2.vy - n1.vy
        d2 = dx ** 2 + dy ** 2

        if d2 < 0.01:
            dx = random() * 0.1 + 0.1
            dy = random() * 0.1 + 0.1
            d2 = dx ** 2 + dy ** 2

        d = sqrt(d2)
        return dx, dy, d

    def _repulse(self, n1, n2):
        dx, dy, d = self._distance(n1, n2)

        if d < self.r:
            f = self.k ** 2 / d ** 2
            n2.force.x += f * dx
            n2.force.y += f * dy
            n1.force.x -= f * dx
            n1.force.y -= f * dy

    def _attract(self, n1, n2, k=0, length=1.0):
        dx, dy, d = self._distance(n1, n2)
        d = min(d, self.r)

        # Take the edge's weight (k) into account.
        f = (d ** 2 - self.k ** 2) / self.k * length
        f *= k * 0.5 + 1
        f /= d

        n2.force.x -= f * dx
        n2.force.y -= f * dy
        n1.force.x += f * dx
        n1.force.y += f * dy



#!/usr/bin/env python3

# This file is part of Shoebot.
# Copyright (C) 2007-2009 the Shoebot authors
# See the COPYING file for the full license text.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#   Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
#   The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
#   WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
#   EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
#   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
#   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
#   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
""" Shoebot console runner """

import argparse
import gettext
import locale
import shlex
import sys

DEFAULT_SERVERPORT = 7777

OUTPUT_EXTENSIONS = (".png", ".svg", ".ps", ".pdf")
APP = "shoebot"
DIR = sys.prefix + "/share/shoebot/locale"

locale.setlocale(locale.LC_ALL, "")
gettext.bindtextdomain(APP, DIR)
gettext.textdomain(APP)
_ = gettext.gettext


def json_arg(s):
    try:
        import json

        d = json.loads(s)
        return d
    except Exception as e:
        error(
            _(
                "Error parsing JSON, remember single quotes OUTSIDE, double QUOTES inside."
            )
        )
        raise e


def error(message):
    """Prints an error message, the help message and quits"""
    global parser
    print((_("Error: ") + message))
    print()
    parser.print_help()
    sys.exit()


def warn(message):
    """Print a warning message"""
    print((_("Warning: ") + message))


def main():
    global parser

    # use ArgumentParser to interpret commandline options
    parser = argparse.ArgumentParser(_("usage: sbot [options] inputfile.bot [args]"))
    parser.add_argument(
        "script", help="Shoebot / Nodebox script to run (filename or code)", nargs="?"
    )

    group = parser.add_argument_group("Input / Output")
    # IO - Output to file
    group.add_argument(
        "-o",
        "--outputfile",
        dest="outputfile",
        help=_(
            "run script and output to image file (accepts .png .svg .pdf and .ps extensions)"
        ),
        metavar="FILE",
    )

    # Shoebot IO - Sockets
    group.add_argument(
        "-s",
        "--socketserver",
        action="store_true",
        dest="socketserver",
        default=False,
        help=_(
            "run a socket server for external control (will run the script in windowed mode)"
        ),
    )
    group.add_argument(
        "-p",
        "--serverport",
        type=int,
        dest="serverport",
        default=DEFAULT_SERVERPORT,
        help=_("set socketserver port to listen for connections (default is 7777)"),
    )

    # IO - Variables
    group.add_argument(
        "-v",
        "--vars",
        dest="vars",
        default=False,
        help=_(
            "Initial variables, in JSON (Note: Single quotes OUTSIDE, double INSIDE) --vars='{\"variable1\": 1}'"
        ),
    )
    # IO - Namespace
    group.add_argument(
        "-ns",
        "--namespace",
        dest="namespace",
        default=None,
        help=_(
            "Initial namespace, in JSON (Note: Single quotes OUTSIDE, double INSIDE) --namespace='{\"variable1\": 1}'"
        ),
    )
    # IO - IDE integration Shell
    group.add_argument(
        "-l",
        "--l",
        dest="shell",
        action="store_true",
        default=False,
        help=_("Simple shell - for IDE interaction"),
    )

    # IO - Passing args to the bot
    group.add_argument("-a", "--args", dest="script_args", help=_("Pass to the bot"))
    group.add_argument("script_args", nargs="?")

    group = parser.add_argument_group("Bot Lifecycle")
    # Bot Lifecycle
    group.add_argument(
        "-r",
        "--repeat",
        type=int,
        dest="repeat",
        default=False,
        help=_("set number of iteration, multiple images will be produced"),
    )

    group = parser.add_argument_group("Window Management")
    group.add_argument(
        "-w",
        "--window",
        action="store_true",
        dest="window",
        help=_("Run script in a GTK window (default to True unless -o is used)"),
    )
    group.add_argument(
        "-f",
        "--fullscreen",
        action="store_true",
        dest="fullscreen",
        default=False,
        help=_("run in fullscreen mode"),
    )
    group.add_argument(
        "-t",
        "--title",
        action="store",
        dest="title",
        default=None,
        help=_("Set window title"),
    )
    group.add_argument(
        "-c",
        "--close",
        action="store_true",
        dest="close",
        default=False,
        help=_("Close window after running bot (use with -r for benchmarking)"),
    )
    group.add_argument(
        "-dv",
        "--disable-vars",
        action="store_true",
        dest="disable_vars",
        default=False,
        help=_("disable the variables pane when in windowed mode."),
    )

    group = parser.add_argument_group("Debugging / Dev flags")
    group.add_argument(
        "-dn",
        "--diagnose",
        action="store_true",
        default=False,
        help=_("Output information for debugging installation / graphics issues."),
    )
    group.add_argument(
        "-dt",
        "--disable-background-thread",
        action="store_true",
        dest="disable_background_thread",
        default=sys.platform == "darwin",
        help=_("disable running bot code in background thread (default on OSX)."),
    )
    group.add_argument(
        "-V",
        "--verbose",
        action="store_true",
        dest="verbose",
        default=False,
        help=_("Show internal shoebot error information in traceback"),
    )

    # get argparse arguments and check for sanity
    args, extra = parser.parse_known_args()

    if args.diagnose:
        from .diagnose import diagnose

        diagnose()
        sys.exit()

    if not args.script:
        error(
            _(
                "Please specify an input script!\n (check /usr/share/shoebot/examples/ for example scripts)"
            )
        )

    if args.vars:
        vars = json_arg(args.vars)
    else:
        vars = None

    if args.namespace:
        namespace = json_arg(args.namespace)
    else:
        namespace = None

    if args.window:
        window = args.window
    else:
        window = not args.outputfile

    from shoebot.__init__ import run  # https://github.com/shoebot/shoebot/issues/206

    success = run(
        src=args.script,
        outputfile=args.outputfile,
        max_iterations=args.repeat or None,
        window=window,
        fullscreen=args.fullscreen,
        title=args.title,
        close_window=args.close,
        server=args.socketserver,
        port=args.serverport,
        show_vars=args.window and args.disable_vars is False,
        vars=vars or None,
        namespace=namespace,
        run_shell=args.shell,
        args=shlex.split(args.script_args or ""),
        verbose=args.verbose,
        background_thread=not args.disable_background_thread,
    )

    # Return errorcode
    return 0 if success else 1


if __name__ == "__main__":
    # Not to be confused with entrypoint from setup.
    sys.exit(main())



#
# transforms.py from old shoebot
#

import sys

from shoebot.core.backend import cairo
from math import sin, cos
from .bezier import BezierPath

TRANSFORMS = ["translate", "scale", "rotate", "skew", "push", "pop"]
CENTER = "center"
CORNER = "corner"
CORNERS = "corners"

import locale, gettext

APP = "shoebot"
DIR = sys.prefix + "/share/shoebot/locale"
locale.setlocale(locale.LC_ALL, "")
gettext.bindtextdomain(APP, DIR)
# gettext.bindtextdomain(APP)
gettext.textdomain(APP)
_ = gettext.gettext


class Transform:
    """
    This class represents a stack of transformations. Supported operations are
    translation, scaling, rotation and skewing.

    """

    def __init__(self, transform=None):
        self.stack = []
        self.set_matrix(transform)

    def set_matrix(self, transform=None):
        if transform is None:
            pass
        elif isinstance(transform, Transform):
            self.append(transform)
        elif isinstance(transform, (list, tuple)):
            matrix = tuple(transform)
            t = cairo.Matrix(*matrix)
            self.append(t)
        elif isinstance(transform, cairo.Matrix):
            self.append(transform)
        else:
            raise ValueError(
                _("Transform: Don't know how to handle transform %s.") % transform
            )

    def translate(self, x, y):
        t = ("translate", x, y)
        self.stack.append(t)

    def scale(self, x, y):
        t = ("scale", x, y)
        self.stack.append(t)

    def rotate(self, a):
        t = ("rotate", a)
        self.stack.append(t)

    def skew(self, x, y):
        t = ("skew", x, y)
        self.stack.append(t)

    def push(self):
        t = ("push",)
        self.stack.append(t)

    def pop(self):
        t = ("pop",)
        self.stack.append(t)

    def append(self, t):
        if isinstance(t, Transform):
            for item in t.stack:
                self.stack.append(item)
        elif isinstance(t, cairo.Matrix):
            self.stack.append(t)
        else:
            raise ValueError(
                _("Transform: Can only append Transforms or Cairo matrices (got %s)")
                % (t)
            )

    def prepend(self, t):
        if isinstance(t, Transform):
            newstack = []
            for item in t.stack:
                newstack.append(item)
            for item in self.stack:
                newstack.append(item)
            self.stack = newstack
        elif isinstance(t, cairo.Matrix):
            self.stack.insert(0, t)
        else:
            raise ValueError(
                _("Transform: Can only append Transforms or Cairo matrices (got %s)")
                % (t)
            )

    def copy(self):
        return self.__class__(self)

    def __iter__(self):
        for value in self.stack:
            yield value

    ### calculates tranformation matrix
    def get_matrix_with_center(self, x, y, mode):
        m = cairo.Matrix()
        centerx = x
        centery = y
        m_archived = []

        for trans in self.stack:
            if isinstance(trans, cairo.Matrix):
                # multiply matrix
                m *= trans
            elif isinstance(trans, tuple) and trans[0] in TRANSFORMS:
                # parse transform command
                cmd = trans[0]
                args = trans[1:]
                t = cairo.Matrix()

                if cmd == "translate":
                    xt = args[0]
                    yt = args[1]
                    m.translate(xt, yt)
                elif cmd == "rotate":
                    if mode == "corner":
                        # apply existing transform to cornerpoint
                        deltax, deltay = m.transform_point(0, 0)
                        a = args[0]
                        ct = cos(a)
                        st = sin(a)
                        m *= cairo.Matrix(
                            ct,
                            st,
                            -st,
                            ct,
                            deltax - (ct * deltax) + (st * deltay),
                            deltay - (st * deltax) - (ct * deltay),
                        )
                    elif mode == "center":
                        # apply existing transform to centerpoint
                        deltax, deltay = m.transform_point(centerx, centery)
                        a = args[0]
                        ct = cos(a)
                        st = sin(a)
                        m *= cairo.Matrix(
                            ct,
                            st,
                            -st,
                            ct,
                            deltax - (ct * deltax) + (st * deltay),
                            deltay - (st * deltax) - (ct * deltay),
                        )
                elif cmd == "scale":
                    if mode == "corner":
                        t.scale(args[0], args[1])
                        m *= t
                    elif mode == "center":
                        # apply existing transform to centerpoint
                        deltax, deltay = m.transform_point(centerx, centery)
                        x, y = args
                        m1 = cairo.Matrix()
                        m2 = cairo.Matrix()
                        m1.translate(-deltax, -deltay)
                        m2.translate(deltax, deltay)
                        m *= m1
                        m *= cairo.Matrix(x, 0, 0, y, 0, 0)
                        m *= m2

                elif cmd == "skew":
                    if mode == "corner":
                        x, y = args
                        ## TODO: x and y should be the tangent of an angle
                        t *= cairo.Matrix(1, 0, x, 1, 0, 0)
                        t *= cairo.Matrix(1, y, 0, 1, 0, 0)
                        m *= t
                    elif mode == "center":
                        # apply existing transform to centerpoint
                        deltax, deltay = m.transform_point(centerx, centery)
                        x, y = args
                        m1 = cairo.Matrix()
                        m2 = cairo.Matrix()
                        m1.translate(-deltax, -deltay)
                        m2.translate(deltax, deltay)
                        t *= m
                        t *= m1
                        t *= cairo.Matrix(1, 0, x, 1, 0, 0)
                        t *= cairo.Matrix(1, y, 0, 1, 0, 0)
                        t *= m2
                        m = t
                elif cmd == "push":
                    m_archived.append(m)
                elif cmd == "pop":
                    m = m_archived.pop()

        return m

    def get_matrix(self):
        """Returns this transform's matrix. Its centerpoint is presumed to be
        (0,0), which is the Cairo default."""
        return self.get_matrix_with_center(0, 0)

    def transformBezierPath(self, path):
        # From nodebox
        if isinstance(path, BezierPath):
            path = path.copy()
        else:
            raise ValueError("Can only transform BezierPaths")

        # for point in path:
        # print(point)
        # path._nsBezierPath = self._nsAffineTransform.transformBezierPath_(path._nsBezierPath)
        return path


class TransformMixin(object):

    """Mixin class for transformation support.
    Adds the _transform and _transformmode attributes to the class."""

    def __init__(self):
        self._reset()

    def _reset(self):
        self._transform = Transform()
        self._transformmode = CENTER

    def _get_transform(self):
        return self._transform

    def _set_transform(self, transform):
        self._transform = Transform(transform)

    transform = property(_get_transform, _set_transform)

    def _get_transformmode(self):
        return self._transformmode

    def _set_transformmode(self, mode):
        self._transformmode = mode

    transformmode = property(_get_transformmode, _set_transformmode)

    def reset(self):
        self._transform = Transform()

    def rotate(self, degrees=0, radians=0):
        self._transform.rotate(degrees, radians)

    def translate(self, x=0, y=0):
        self._transform.translate(x, y)

    def scale(self, x=1, y=None):
        self._transform.scale(x, y)

    def skew(self, x=0, y=0):
        self._transform.skew(x, y)



import argparse
import subprocess
import tempfile
from pathlib import Path


def main():
    global parser
    parser = argparse.ArgumentParser("usage: sbot [options] inputfile.bot [args]")
    parser.add_argument("script", help="Shoebot script to run")

    parser.add_argument(
        "-o",
        "--outputfile",
        dest="outputfile",
        help="destination file (.mp4 only)",
        metavar="FILE",
    )
    parser.add_argument(
        "-f",
        "--frames",
        dest="framenumber",
        default=300,
        help="number of frames to export (default 300)",
    )

    args, extra = parser.parse_known_args()

    if not args.outputfile:
        outfile = args.script.replace(".bot", ".mp4")
    else:
        outfile = args.outputfile

    with tempfile.TemporaryDirectory() as tmpdirname:
        tmpdirpath = Path(tmpdirname)

        image_name = tmpdirpath / Path(args.script).with_suffix(".png").name
        cmd = f"sbot {args.script} --repeat {args.framenumber} \
            --outputfile {image_name}"
        result = subprocess.call(cmd, shell=True)
        if result != 0:
            sys.exit(1)

        fileglob = tmpdirpath / "*.png"
        cmd = f"ffmpeg -y -loglevel 24 -r 30 -f image2 -pattern_type glob \
          -i '{fileglob}' -c:v libx264 -crf 20 -movflags faststart -c:a aac \
          -pix_fmt yuv420p {outfile}"
        subprocess.call(cmd, shell=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Shoebot setup script
#
# python setup.py install', or
#    python setup.py --help' for more options
#

import glob
import itertools
import os
import platform
import re
import shutil
import sys
import textwrap

info = textwrap.dedent(
    """
    Shoebot is a pure Python graphics robot: It takes a Python script as input,
    which describes a drawing process, and outputs a graphic in a common open
    standard format (SVG, PDF, PostScript, or PNG). It has a simple text editor
    GUI, and scripts can describe their own GUIs for controlling variables
    interactively. Being pure Python, it can also be used as a Python module,
    a plugin for Python-scriptable tools such as Inkscape, and run from the
    command line.
"""
)

if sys.version_info < (3, 4):
    sys.exit("Shoebot requires python 3.4 or higher.")

# the following libraries will not be installed
EXCLUDE_LIBS = ["lib/sbopencv", "lib/sbopencv/blobs"]

is_pypy = "__pypy__" in sys.builtin_module_names
is_jython = platform.system == "Java"

here = os.path.dirname(os.path.abspath(__file__))

try:
    from setuptools import find_packages, setup, Command
except ImportError:
    sys.exit("Install setuptools before shoebot")


class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""

    CLEAN_FILES = "./build ./dist ./*.pyc ./*.tgz ./*.egg-info ./.eggs".split()

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        global here

        for path_spec in self.CLEAN_FILES:
            # Make paths absolute and relative to this path
            abs_paths = glob.glob(os.path.normpath(os.path.join(here, path_spec)))
            for path in [str(p) for p in abs_paths]:
                if not path.startswith(here):
                    # Die if path in CLEAN_FILES is absolute + outside this directory
                    raise ValueError("%s is not a path inside %s" % (path, here))
                print("removing %s" % os.path.relpath(path))
                shutil.rmtree(path)


class DiagnoseCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from shoebot.diagnose import diagnose

        diagnose()


# dir globbing approach taken from Mercurial's setup.py
datafiles = [
    (
        os.path.join("share/shoebot/", root),
        [os.path.join(root, file_) for file_ in files],
    )
    for root, dir, files in itertools.chain(os.walk("examples"), os.walk("locale"))
]
datafiles.extend(
    [
        ("share/pixmaps", ["assets/shoebot-ide.png"]),
        ("share/shoebot/data", ["assets/kant.xml"]),
        ("share/applications", ["assets/shoebot-ide.desktop"]),
    ]
)

# include all libs EXCEPT the ones mentioned in EXCLUDE_LIBS
for lib in EXCLUDE_LIBS:
    # get subdirs of excluded libs
    for root, dir, files in list(os.walk(lib))[1:]:
        EXCLUDE_LIBS.append(root)

datafiles.extend(
    [
        (
            os.path.join("share/shoebot/", root),
            [os.path.join(root, file_) for file_ in files],
        )
        for root, dir, files in os.walk("lib")
        if root not in EXCLUDE_LIBS
    ]
)

PYCAIRO = "pycairo>=1.18.1"
PYGOBJECT = "pygobject>=3.32.1"
# Also requires one of 'vext.gi' or 'pgi' to run in GUI
BASE_REQUIREMENTS = [
    "setuptools>=18.8",
    PYCAIRO,
    "Pillow>=6.0.0",
    "pubsub==0.1.2",
    "wrapt>=1.12.1",
]


def read_version():
    global here
    with open(f"{here}/VERSION", mode="r") as f:
        content = f.read()
        version_info = dict(re.findall(r"(^[a-z]+)\s+(.+)$", content, re.MULTILINE))
    return version_info["version"]


def requirements(debug=True, with_pgi=None):
    """
    Build requirements based on flags

    :param with_pgi: Use 'pgi' instead of 'gi' - False on CPython, True elsewhere
    :return:
    """
    reqs = list(BASE_REQUIREMENTS)
    if with_pgi is None:
        with_pgi = is_jython

    if debug:
        print("setup options: ")
        print("with_pgi:      ", "yes" if with_pgi else "no")
    if with_pgi:
        reqs.append("pgi")
        if debug:
            print("warning, as of April 2019 typography does not work with pgi")
    else:
        reqs.append(PYGOBJECT)

    if debug:
        print("")
        print("")
        for req in reqs:
            print(req)
    return reqs


setup(
    name="shoebot",
    version=read_version(),
    description="Vector graphics scripting application",
    long_description=info,
    author="Ricardo Lafuente",
    author_email="r@manufacturaindependente.org",
    license="GPL v3",
    url="http://shoebot.net",
    cmdclass={"clean": CleanCommand, "diagnose": DiagnoseCommand},
    packages=find_packages(exclude=["tests*", "extensions"]),
    data_files=datafiles,
    setup_requires=[PYCAIRO],
    install_requires=requirements(
        debug="install" in sys.argv,
        with_pgi=os.environ.get("SHOEBOT_GI", False) == "pgi",
    ),
    tests_require=["parameterized", "pyyaml", "wrapt"],
    entry_points={
        "console_scripts": [
            "sbot=shoebot.run:main",
            "shoebot=shoebot.ide.ide:main",
            "sbot-video-export=shoebot.util.video_export:main",
        ],
        "gui_scripts": "shoebot=shoebot.ide.ide:main",
    },
    test_suite="tests.unittests",
)



"""
Simple command shell

Launch shoebot with -l to activate it.

IDEs can launch shoebot with -l and communicate via this shell
You can also try it as a user:

$ sbot -l ~/examples/animation/hypnoval.bot


For live coding, IDEs can use the load_base64 command to load new
source:  The first parameter is the base64 encoded source code.

Cookie parameter:

If an editor wants to know the status of specific commands, set
cookie=unique_value as the last parameter of any command.

The cookie will be the first text on the line, followed by > or :

Examples:

%cookie> this is an intermediate line
%cookie>
%cookie: this is the last line, the client can dispose of the cookie

%cookie status>
%cookie:

%cookie status:

This makes it possible to filter responses from other info on stdout.


List available commands:

Type 'help' to list all the available commands.
"""

from __future__ import print_function
import base64
import cmd
import shlex

from shoebot.core.events import (
    QUIT_EVENT,
    SOURCE_CHANGED_EVENT,
    publish_event,
    SET_WINDOW_TITLE_EVENT,
)

PROMPT = ""
RESPONSE_PROMPT = ""
INTRO = RESPONSE_PROMPT + '"Shoebot Shell."'

RESPONSE_CODE_OK = "CODE_OK"
RESPONSE_REVERTED = "REVERTED"

trusted_cmds = set()


def trusted_cmd(f):
    """
    Decorator for methods that are trusted commands.

    If self.trusted is not set, then the command prints an error.

    :param f: do_XXX function to decorate.
    """

    def run_cmd(self, line):
        if self.trusted:
            f(self, line)
        else:
            print("Sorry cannot do %s here." % f.__name__[3:])

    global trusted_cmds
    trusted_cmds.add(f.__name__)
    run_cmd.__doc__ = f.__doc__
    return run_cmd


class ShoebotCmd(cmd.Cmd):
    """Simple command processor example."""

    shortcuts = {
        "rw": "rewind",
        "?": "help",
        "h": "help",
        "s": "speed",
        "g": "goto",
        "p": "pause",
        "q": "quit",
        "r": "restart",
    }

    # Trusted commands can only be run if trusted is set, e.g. when running from the commandline
    # not a socket.
    trusted_cmds = set()

    def __init__(self, bot, intro=None, trusted=False, **kwargs):
        """

        :param bot: bot instance
        :param intro: intro banner to display.
        :param trusted: Only running from the commandline is trusted, not from sockets
                        untrusted can only change variables
        :param kwargs:
        :return:
        """
        cmd.Cmd.__init__(self, **kwargs)
        self.bot = bot
        self.pause_speed = None
        self.intro = intro or INTRO
        self.prompt = PROMPT
        self.response_prompt = ""
        self.use_rawinput = False
        self.cookie = None
        self.escape_nl = False
        self.live_prefix = ""
        self.trusted = trusted

    def print_response(self, input="", keep=False, *args, **kwargs):
        """
        print response, if cookie is set then print that each line
        :param args:
        :param keep: if True more output is to follow.
        :param cookie: set a custom cookie,
                       if set to 'None' then self.cookie will be used.
                       if set to 'False' disables cookie output entirely
        """
        status = kwargs.get("status")
        lines = input.splitlines()

        # Get cookie - unique string that can be provided by the sender to
        # filter replies to its requests that can arrive later.
        cookie = kwargs.get("cookie")
        if cookie is None:
            cookie = self.cookie or ""

        if status and not lines:
            # Nothing to output, just send a newline.
            print("", file=self.stdout)
            return

        if not cookie:
            # Standard output, lines arrive as they are.
            for line in lines:
                print(line, file=self.stdout)
            return

        # Prefix lines with unique str "cookie" and : or > on the last line.
        for i, line in enumerate(lines):
            if i != len(lines) - 1 or keep is True:
                cookie_char = ">"
            else:
                # last line
                cookie_char = ":"

            print(
                f"{cookie} {status or ''}{cookie_char}{line.strip()}",
                file=self.stdout,
            )

    def emptyline(self):
        """
        Override the default behaviour of repeating the last line.

        :return:
        """
        return ""

    def do_escape_nl(self, arg):
        """
        Toggle escaping newlines.
        """
        if arg.lower() == "off":
            self.escape_nl = False
        else:
            self.escape_nl = True

    def do_prompt(self, arg):
        """
        Enable or disable prompt
        :param arg: on|off
        :return:
        """
        if arg.lower() == "off":
            self.response_prompt = ""
            self.prompt = ""
            return
        elif arg.lower() == "on":
            self.prompt = PROMPT
            self.response_prompt = RESPONSE_PROMPT
        self.print_response(
            f"prompt: {self.prompt}\n" + f"response: {self.response_prompt}"
        )

    def do_title(self, title):
        """
        Change window title.
        """
        publish_event(SET_WINDOW_TITLE_EVENT, data=title)

    def do_speed(self, speed):
        """
        rewind
        """
        if speed:
            try:
                new_speed = float(speed)
            except ValueError as e:
                self.print_response(f"{speed} is not a valid framerate")
                return

            self.bot._speed = new_speed
        self.print_response(f"Speed: {self.bot._speed} FPS")

    def do_restart(self, line):
        """
        Attempt to restart bot by clearing its namespace and resetting FRAME to 0.
        """
        self.bot._frame = 0
        self.bot._namespace.clear()  # noqa
        self.bot._namespace.update(self.bot._initial_namespace)  # noqa

    def do_pause(self, line):
        """
        Toggle paused state.

        Stores the bots speed in pause_speed and temporarily sets the bot speed to 0.
        """
        if self.pause_speed is None:
            self.pause_speed = self.bot._speed
            self.bot._speed = 0
            self.print_response("Paused")
            return

        self.bot._speed = self.pause_speed
        self.pause_speed = None
        self.print_response("Playing")

    def do_play(self, line):
        """
        Resume playback if bot is paused
        """
        if self.pause_speed is not None:
            self.bot._speed = self.pause_speed
            self.pause_speed = None
        self.print_response("Play")

    def do_goto(self, line):
        """
        Go to specific frame
        """
        try:
            new_frame = int(line)
        except ValueError:
            self.print_response(f"{line} is not a valid framerate")
            return

        self.print_response(f"Go to frame {new_frame}")
        self.bot._frame = new_frame

    def do_rewind(self, line):
        """
        rewind
        """
        self.print_response(f"Rewinding from frame {self.bot._frame} to 0")
        self.bot._frame = 0

    def do_vars(self, line):
        """
        List bot variables and values.
        """
        if self.bot._vars:  # noqa
            max_name_len = max([len(name) for name in self.bot._vars])  # noqa
            for i, (name, v) in enumerate(self.bot._vars.items()):  # noqa
                keep = i < len(self.bot._vars) - 1  # noqa
                self.print_response(
                    f"{name.ljust(max_name_len)} = {v.value}", keep=keep
                )
        else:
            self.print_response("No vars")

    @trusted_cmd
    def do_load_base64(self, line):
        """
        load filename=(file)
        load base64=(base64 encoded)

        Send new code to shoebot.

        If it does not run successfully shoebot will attempt to role back.

        Editors can enable live-coding by sending new code as it is edited.
        """
        # TODO use publish_event to change source, and let the mainloop update the executor.
        cookie = self.cookie
        executor = self.bot._executor  # noqa

        def source_good():
            self.print_response(status=RESPONSE_CODE_OK, cookie=cookie)
            executor.clear_callbacks()

        def source_bad(tb):
            if called_good:
                # good and bad callbacks shouldn't both be called
                raise ValueError(
                    "Unexpected condition, Good and bad callbacks were called !"
                )

            self.print_response(status=RESPONSE_REVERTED, keep=True, cookie=cookie)
            self.print_response(tb.replace("\n", "\\n"), cookie=cookie)
            executor.clear_callbacks()

        called_good = False
        source = base64.b64decode(line).decode("utf-8")
        # Test compile
        publish_event(
            SOURCE_CHANGED_EVENT, data=source, extra_channels="shoebot.source"
        )
        self.bot._executor.load_edited_source(  # noqa
            source, good_cb=source_good, bad_cb=source_bad
        )

    def do_bye(self, line):
        """
        Exit shell and shoebot

        Alias for exit.
        """
        return self.do_exit(line)

    def do_exit(self, line):
        """
        Exit shell and shoebot
        """
        if self.trusted:
            publish_event(QUIT_EVENT)
        self.print_response("Bye.\n")
        return True

    def do_quit(self, line):
        """
        Exit shell and shoebot

        Alias for exit.
        """
        return self.do_exit(line)

    def do_fullscreen(self, line):
        """
        Make the current window fullscreen
        """
        # TODO: use publish_event to set toggle fullscreen instead of calling trigger_fullscreen_action directly.
        self.bot.canvas.sink.trigger_fullscreen_action(True)
        print(self.response_prompt, file=self.stdout)

    def do_window(self, line):
        """
        Un-fullscreen and
        """
        # TODO: use publish_event to set toggle fullscreen instead of calling trigger_fullscreen_action directly.
        self.bot.canvas.sink.trigger_fullscreen_action(False)
        print(self.response_prompt, file=self.stdout)

    def do_EOF(self, line):
        """
        Exit shell and shoebot

        Alias for exit.
        """
        print(self.response_prompt, file=self.stdout)
        return self.do_exit(line)

    def do_help(self, arg):
        """
        Show help on all commands.
        """
        print(self.response_prompt, file=self.stdout)
        return cmd.Cmd.do_help(self, arg)

    def do_set(self, line):
        """
        Set a variable.
        """
        # TODO: use publish_event to set Variable instead of directly changing it.
        try:
            name, value = [part.strip() for part in line.split("=")]
            if name not in self.bot._vars:  # noqa
                self.print_response(
                    "No such variable %s enter vars to see available vars" % name
                )
                return
            variable = self.bot._vars[name]
            variable.value = variable.sanitize(value.strip(";"))

            success, msg = self.bot.canvas.sink.var_changed(name, variable.value)
            if success:
                print(f"{name}={variable.value}", file=self.stdout)
            else:
                print(f"{msg}\n", file=self.stdout)
        except Exception:
            print("Invalid Syntax for set command")

    def precmd(self, line):
        """
        Preprocess the commandline.

        Cookies:
            Allow commands to have a last parameter of 'cookie=arbitrary_string'

            Editors can send a "cookie" value - an arbitrary string, usually
            a uuid, for commands that have responses, this will be prepended
            to the response, along with : if there is more data to follow
            or > on the last line.

        :param line:  input command line.
        :return:  processed command line.
        """
        args = shlex.split(line or "")
        if not args or args[0].startswith("#"):
            # Ignore empty lines and comments.
            return ""

        if "cookie=" in args[-1]:
            # Get the cookie value.
            line, self.cookie = line.rsplit("cookie=")
            line = line.strip()
            del args[-1]
            if not args:
                return ""

        if hasattr(self, f"do_{args[0]}"):
            # This stops var=value parsing trying to handle load_base64 which
            # may contain = signs.
            return line

        # Further parsing
        if "=" in line:
            # Variable Assignment: somevar=value
            #
            # Convert var=value into set var=value
            #
            if "set" not in args[0]:
                return f"set {line}"
            return line
        elif args[0] in self.shortcuts:
            # Shortcuts
            #
            # If the first item on the commandline is a shortcut, the substitute it for
            # the full command.
            return self.shortcuts[args[0]] + " ".join(args[1:])

        return line

    def postcmd(self, stop, line):
        """Hook method executed just after a command dispatch is finished."""
        self.cookie = None
        return stop

    def postloop(self):
        print("", file=self.stdout)



from __future__ import print_function

import copy
import dataclasses
import os
import sys
import traceback
from math import copysign
from time import sleep, time

from .livecode import LiveExecution
from shoebot.core.events import (
    event_is,
    next_event,
    QUIT_EVENT,
    SET_WINDOW_TITLE_EVENT,
    SOURCE_CHANGED_EVENT,
    VARIABLE_CHANGED_EVENT,
    REDRAW_EVENT,
)
from shoebot.core.var_listener import VarListener
from shoebot.data import Variable
from shoebot.grammar.format_traceback import simple_traceback
from shoebot.util import UnbufferedFile

sys.stdout = UnbufferedFile(sys.stdout)
sys.stderr = UnbufferedFile(sys.stderr)

# If a bot is not animated the GUI is updated at 30fps
DEFAULT_GUI_UPDATE_SPEED = 30.0
DEFAULT_ANIMATION_SPEED = 60.0


class Grammar(object):
    """
    A Bot is an interface to receive user commands (through scripts or direct
    calls) and pass them to a canvas for drawing.

    Bae class for all Grammars, contains just the machinery for running the
    grammars, it has only the private API and nothing else, except for
    run which is called to actually run the Bot.
    """

    def __init__(self, canvas, namespace=None, vars=None):
        self._canvas = canvas
        self._dynamic = True  ##
        self._speed = None
        self._vars = vars or {}
        self._oldvars = self._vars
        self._namespace = namespace or {}

        input_device = canvas.get_input_device()
        if input_device:
            input_device.set_callbacks(
                key_pressed=self._key_pressed,
                key_released=self._key_released,
                mouse_button_down=self._mouse_button_down,
                mouse_button_up=self._mouse_button_up,
                mouse_pointer_moved=self._mouse_pointer_moved,
            )
        self._input_device = input_device

    def _update_animation_variables(self, iteration, frame):
        """
        :param iteration: Increases by one every time the bot code is executed.
        :param frame: If speed > 0 increases by one each time the bot code is executed.
                      If speed is negative then decreases each time the bot code is executed.
        """
        # Update bot variables that change on each animation frame.
        self._namespace["ITERATION"] = iteration
        self._namespace["FRAME"] = frame
        self._namespace["PAGENUM"] = frame

    def _load_namespace(self, namespace, filename=None):
        """
        Initialise bot namespace with info in shoebot.data

        :param filename: Will be set to __file__ in the namespace
        """
        from shoebot import data

        for name in dir(data):
            namespace[name] = getattr(data, name)

        for name in dir(self):
            if name[0] != "_":
                namespace[name] = getattr(self, name)

        namespace["_ctx"] = self  # Used in older nodebox scripts.
        namespace["__file__"] = filename

    #### Execute a single frame

    def _calculate_frame_delay(self, speed, start_time):
        """
        :return in seconds time to delay, taking into account execution time
        """
        if not speed:
            return 0

        # If the amount of time taken is more than the FPS delay
        # then return zero.
        return max((1.0 / abs(speed)) - (time() - start_time), 0.0)

    def run(
        self,
        inputcode,
        max_iterations=None,
        run_forever=False,
        frame_limiter=False,
        verbose=False,
    ):
        if os.path.isfile(inputcode):
            source = open(inputcode).read()
            filename = inputcode
        elif isinstance(inputcode, str):
            filename = "<string>"
            source = inputcode
        else:
            raise ValueError("inputcode must be a str or file like object.")

        self._load_namespace(self._namespace, filename)
        # TODO:  The shell module (sbio) accesses the executor via its name here,
        # making this event based would remove the need for this.
        self._executor = executor = LiveExecution(
            source, ns=self._namespace, filename=filename
        )

        if run_forever is False:
            if max_iterations is None:
                max_iterations = 1

        try:
            # Iterations only increment, whereas FRAME can decrement if the user sets a negative speed.
            iteration = 0
            first_run = True
            while first_run or iteration != max_iterations:
                # Main loop:
                # - Setup bot on first run.
                # - Run draw function for if present.
                # - Process events
                # - Update state
                start_time = time()
                iteration += 1

                canvas_dirty = False
                # Reset output graphics state
                self._canvas.reset_canvas()

                with executor.run_context() as (known_good, source, ns):
                    if not known_good:
                        # New code has been loaded, but it may have errors.
                        # Setting first_run forces the global context to be re-run
                        # Which has the side effect of loading all functions and state.
                        first_run = True

                    if first_run:
                        # Run code in the global namespace, followed by setup()
                        executor.run()
                        if "setup" in executor.ns:
                            executor.ns["setup"]()

                        if "draw" in executor.ns:
                            if self._speed is None:
                                self._speed = DEFAULT_ANIMATION_SPEED

                        # Store initial state so script can revert to a known state when livecoding.
                        self._initial_namespace = copy.copy(self._namespace)
                        canvas_dirty = True

                    is_animation = "draw" in executor.ns
                    if is_animation and self._speed != 0:
                        # If speed is 0, then don't output anything..
                        executor.ns["draw"]()
                        canvas_dirty = True

                if canvas_dirty:
                    self._canvas.flush(self._frame)

                if frame_limiter:
                    # Frame limiting is only used when running the GUI.
                    if is_animation:
                        # User specifies framerate, via speed(...) or use a default.
                        fps = self._speed
                        timeout = self._calculate_frame_delay(
                            fps if fps is not None else DEFAULT_ANIMATION_SPEED,
                            start_time,
                        )
                        next_frame_due = time() + timeout
                    else:
                        # Re-run the mainloop at 30fps, so that the GUI remains responsive.
                        next_frame_due = time() + 1.0 / DEFAULT_GUI_UPDATE_SPEED
                else:
                    # Do not sleep between frames.
                    next_frame_due = time()

                # Handle events
                continue_running, first_run = self._handle_events(
                    iteration, is_animation, next_frame_due
                )
                if not continue_running:
                    # Event handler returns False if it receives a message to quit.
                    break

            # Main loop has finished, return True to indicate it exited normally.
            return True
        except Exception as e:
            # Catch Exception, not BaseException, so that KeyboardInterrupts (ctrl+c) still work.
            # if something goes wrong, print verbose system output.

            import sys

            if verbose:
                errmsg = traceback.format_exc()
            else:
                errmsg = simple_traceback(e, executor.known_good or "")
            sys.stderr.write(f"{errmsg}\n")
            return False

    def _handle_events(self, iteration, is_animation, next_frame_due):
        """
        The Shoebot mainloop, GUI and shell communicate with each other using events.

        Examples include live variables being changed from the GUI, the shell
        or Shoebot itself, or the user quitting from the GUI.

        This handler waits for events and updates where needed, the loop also
        serves handles the delay between frames for animated bots.

        return: continue_running, restart
        """

        # Things we might want to do on returning:
        # Restart (if state has changed and not an animation).
        # Quit
        # Continue running.

        restart_bot = False
        while True:
            timeout = min(next_frame_due - time(), 0.1)
            event = next_event(
                block=timeout > 0, timeout=timeout if timeout > 0 else None
            )
            # Update GUI, which may in-turn generate new events.
            self._canvas.sink.main_iteration()

            if event is not None:
                if event.type == QUIT_EVENT:
                    # The user chose to quit via the shell or GUI.
                    return False, False
                elif event.type == REDRAW_EVENT:
                    # The GUI needs redrawing (usually because the Window was resized)
                    # TODO: This is a hack/workaround, since the graphics backend doesn't currently support redrawing
                    if not is_animation:
                        return True, True
                elif event.type == SET_WINDOW_TITLE_EVENT:
                    # A new window title was specified in the shell
                    self._canvas.sink.set_title(event.data)
                elif event.type == SOURCE_CHANGED_EVENT:
                    # New source code was loaded from the shell.
                    # Debounce SOURCE_CHANGED events -
                    # Gedit generates two events for changing a single character -
                    # delete and then add
                    while event and event.type == SOURCE_CHANGED_EVENT:
                        # TODO, can this be handled differently (non-blocking or just ignore source that is the same?)
                        event = next_event(block=True, timeout=0.001)
                    if not is_animation:
                        return True, True
                elif event.type == VARIABLE_CHANGED_EVENT:
                    # A Variable was changed, from the shell or the GUI.
                    # TODO, make VARIABLE_ADDED_EVENT, VARIABLE_DELETED_EVENT
                    # TODO, sketched out, fix up properly.
                    self._executor.ns[event.data.name] = event.data.value
                    # TODO: State was updated, bot needs to execute again ???
                    if not is_animation:
                        # On non-animated bots, updating variables re-runs the whole
                        # whole bot so that the user may see the updated state.
                        return True, True

            if time() >= next_frame_due:
                break

        if event is None:
            # event is None indicates the handler timed out.
            # If the bot is animated, then the next frame is due and
            # variables that update per-frame must be updated.

            if is_animation and self._speed is not None:
                if self._speed > 0:
                    self._frame += 1
                elif self._speed < 0:
                    self._frame -= 1
            self._update_animation_variables(iteration, self._frame)

        # By default return continue_running=True.
        return True, restart_bot

    def finish(self):
        ## For use when using shoebot as a module
        ## TODO: Not used when running as a bot, possibly should not be available in
        ## this case
        self._canvas.flush(self._frame)

    #### Variables
    def _addvar(self, v):
        """Sets a new accessible variable.

        :param v: Variable.
        """
        oldvar = self._oldvars.get(v.name)
        if oldvar is not None:
            if isinstance(oldvar, Variable):
                if oldvar.compliesTo(v):
                    v.value = oldvar.value
            else:
                # Set from commandline
                v.value = v.sanitize(oldvar)
        else:
            for listener in VarListener.listeners:
                listener.var_added(v)
        self._vars[v.name] = v
        self._namespace[v.name] = v.value
        self._oldvars[v.name] = v
        return v

    def _findvar(self, name):
        for v in self._oldvars:
            if v.name == name:
                return v
        return None

    #### For override by implementing grammars: ####
    def _mouse_button_down(self, button):
        pass

    def _mouse_button_up(self, button):
        pass

    def _mouse_pointer_moved(self, x, y):
        pass

    def _key_pressed(self, key, keycode):
        pass

    def _key_released(self, key, keycode):
        pass



#!/usr/bin/env python3
import gi

gi.require_version("PangoCairo", "1.0")
from gi.repository import PangoCairo

# https://developer.gnome.org/pygtk/stable/index.html


def fontnames():
    """Returns a list of available Pango font names. Names are returned in the
    Pango font string format, e.g. "Inconsolata Bold"."""
    names = []
    families = PangoCairo.font_map_get_default().list_families()
    for family in families:
        # FIXME: Ensure this is the correct way to make a font name
        facenames = [
            f"{family.get_name()} {face.get_face_name()}"
            for face in family.list_faces()
        ]
        names.extend(facenames)
        # description = face.describe()
    return names


if __name__ == "__main__":
    print(fontnames())



import _ast
import ast
import copy
import contextlib
import threading
import traceback


class LiveExecution(object):
    """
    Live Code executor.

    Code has two states, 'known good' and 'tenous'

    Known good:  Exceptions are raised as normal
    Tenuous: An exception will cause the code to be reverted to the last Known Good code

    Initially code is known-good, new code sent to the executor is tenuous, until
    it has been executed at least once.
    """

    ns = {}
    lock = threading.RLock()

    def __init__(self, source, ns=None, filename=None):
        self.edited_source = None
        self.known_good = source
        self.filename = filename
        if ns is None:
            self.ns = {}
        else:
            self.ns = ns

        self.good_cb = None
        self.bad_cb = None

    @property
    def is_edited(self):
        """
        :return: True if source has been edited
        """
        return self.edited_source is not None

    def load_edited_source(self, source, good_cb=None, bad_cb=None, filename=None):
        """
        Load changed code into the execution environment.

        Until the code is executed correctly, it will be
        in the 'tenuous' state.
        """
        with LiveExecution.lock:
            self.good_cb = good_cb
            self.bad_cb = bad_cb
            try:
                # Test compile (TODO - keep the binary object + use that)
                source = f"{source}\n\n"
                compile(source, filename or self.filename, "exec")
                self.edited_source = source
            except Exception as e:  # noqa
                if bad_cb:
                    self.edited_source = None
                    tb = traceback.format_exc()
                    self.call_bad_cb(tb)
                return
            if filename is not None:
                self.filename = filename

    def reload_functions(self, edited=True):
        """
        Reload functions, from edited_source or known_good

        :param edited: if True then functions are reloaded from edited_source, otherwise known_good is used.
        """
        if edited:
            source = self.edited_source
        else:
            source = self.known_good

        with LiveExecution.lock:
            if source is not None:
                compiled_modules = {}
                tree = ast.parse(source)
                for f in [n for n in ast.walk(tree) if isinstance(n, ast.FunctionDef)]:
                    module = _ast.Module(body=[f], type_ignores=[])
                    code = compiled_modules.get(module)
                    if code is None:
                        code = compile(module, self.filename, "exec")
                        compiled_modules[module] = code

                    function_context = {}
                    eval(code, dict(self.ns), function_context)
                    self.ns[f.name].__code__ = function_context[f.name].__code__

    def do_exec(self, source, ns):
        """
        Override if you want to do something other than exec in ns

        tenuous is True if the source has just been edited and may fail
        """
        exec(source, ns)

    def run_tenuous(self):
        """
        Run edited source, if no exceptions occur then it
        graduates to known good.
        """
        with LiveExecution.lock:
            ns_snapshot = copy.copy(self.ns)
            try:
                source = self.edited_source
                self.edited_source = None
                self.do_exec(source, ns_snapshot)
                self.known_good = source
                self.call_good_cb()
                return True, None
            except Exception as ex:
                tb = traceback.format_exc()
                self.call_bad_cb(tb)
                self.ns.clear()
                self.ns.update(ns_snapshot)
                return False, ex

    def run(self):
        """
        Attempt to known good or tenuous source.
        """
        with LiveExecution.lock:
            if self.edited_source:
                success, ex = self.run_tenuous()
                if success:
                    return

            self.do_exec(self.known_good, self.ns)

    def clear_callbacks(self):
        """
        clear the good and bad callbacks
        """
        with LiveExecution.lock:
            self.bad_cb = None
            self.good_cb = None

    def call_bad_cb(self, tb):
        """
        If bad_cb returns True then keep it
        :param tb: traceback that caused exception
        :return:
        """
        with LiveExecution.lock:
            if self.bad_cb and not self.bad_cb(tb):
                self.bad_cb = None

    def call_good_cb(self):
        """
        If good_cb returns True then keep it
        :return:
        """
        with LiveExecution.lock:
            if self.good_cb and not self.good_cb():
                self.good_cb = None

    @contextlib.contextmanager
    def run_context(self):
        """
        Context in which the user can run the source in a custom manner.

        If no exceptions occur then the source will move from 'tenuous'
        to 'known good'.

        >>> with run_context() as (known_good, source, ns):
        >>> ...  exec source in ns
        >>> ...  ns['draw']()

        """
        with LiveExecution.lock:
            if self.edited_source is None:
                yield True, self.known_good, self.ns
                return

            ns_snapshot = copy.copy(self.ns)
            try:
                yield False, self.edited_source, self.ns
                self.known_good = self.edited_source
                self.edited_source = None
                self.call_good_cb()
                return
            except Exception as ex:
                tb = traceback.format_exc()
                self.call_bad_cb(tb)
                self.edited_source = None
                self.ns.clear()
                self.ns.update(ns_snapshot)
                self.reload_functions(edited=False)
